# $Id: emen2resource.py,v 1.33 2011/05/02 16:22:34 edwlan Exp $
import urllib2
import re
import time
import collections

import twisted.internet
import twisted.web.static
import twisted.web.resource

# emen2 imports
import emen2.db.exceptions
import emen2.util.jsonutil
import emen2.util.loganalyzer
import emen2.db.config
from emen2.web import routing
from emen2.web import responsecodes
g = emen2.db.config.g()

class EMEN2Resource(twisted.web.resource.Resource):

	resourcename = 'emen2'

	isLeaf = True

	contentTypes = twisted.web.static.loadMimeTypes()
	contentEncodings = {
			".gz" : "gzip",
			".bz2": "bzip2"
			}

	mimetype = None
	defaultType = "text/html"


	# Anything in a POST body will be treated as JSON.
	def parse_args(self, request):
		ret = {}

		for k,v in request.args.items():
			if len(v) == 1:
				v = v.pop()
			ret[k] = v

		# Process "dictname.keyname" items into dictionaries.
		for k in filter(lambda x:"." in x, ret.iterkeys()):
			dictname, _, keyname = k.partition(".")
			if not ret.get(dictname):
				ret[dictname] = {}
			ret[dictname][keyname] = ret.get(k)
			del ret[k]

		ret.update(self.parse_content(request))

		# These aren't allowed:
		for i in set(["db", "host"]):
			ret.pop(i, None)

		for k in ret.keys():
			kp = k.partition("___")
			if kp[2] == "json":
				ret[kp[0]] = emen2.util.jsonutil.decode(ret[k])
				del ret[k]

		return ret


	def parse_content(self, request):
		# Check for any JSON-encoded data in the POST body
		postargs = {}
		if request.method == "POST":
			request.content.seek(0)
			content = request.content.read()
			if content:
				try:
					postargs = emen2.util.jsonutil.decode(content)
				except:
					pass

		return postargs


	# ian: changed from render_GET to render to accept both GET/POST
	def render(self, request):
		args = self.parse_args(request)
		host = request.getClientIP()
		request.postpath = filter(bool, request.postpath)
		ctxid = request.getCookie("ctxid") or args.get('ctxid')

		t = time.time()
		d = twisted.internet.threads.deferToThread(self._action, request=request, args=args, ctxid=ctxid, host=host)
		d.addCallback(self._cb, request, t=t)
		d.addErrback(self._eb, request, t=t)

		return twisted.web.static.server.NOT_DONE_YET



	def _action(self, path=None, postpath=None, args=None, ctxid=None, host=None, method=None, db=None, request=None):
		# Return a result string, and any headers to set.
		return 'Test!', {}



	def _cb(self, result, request, t=0):
		# Set response headers
		headers = {
			"Content-Type": "text/html; charset=utf-8",
			"X-Username": "",
		}
		#headers.update({"Cache-Control":"no-cache", "Pragma":"no-cache"})

		result, content_headers = result

		headers.update(content_headers)
		headers = dict( (k,v) for k,v in headers.iteritems() if v != None )
		if result is not None:
			headers['Content-Length'] = len(result)

		# Redirect if necessary
		# and request.code != 200:
		if headers.get("Location"):
			request.setResponseCode(302)

		[request.setHeader(key, str(headers[key])) for key in headers]

		# Set the session ctxid
		setctxid = headers.get('X-ctxid')
		if setctxid != None:
			request.addCookie("ctxid", setctxid, path='/')



		# send result to client
		if result is not None:
			request.write(result)
			request.finish()

		#self._logline(request, t=t)



	def _eb(self, failure, request, t=0):
		e = ''
		data = ''
		headers = {}
		response = 500

		g.error(failure)
		# Raise exception
		try:
			if isinstance(failure, BaseException):
				raise
			else:
				failure.raiseException()

		# Authentication exceptions
		except (emen2.db.exceptions.AuthenticationError, emen2.db.exceptions.DisabledUserError, emen2.db.exceptions.SecurityError), e:
			response = 401
			data = self.render_security_error(request.uri, e)

		except emen2.db.exceptions.SessionError, e:
			response = 401
			data = self.render_security_error(request.uri, e)
			# dead sessions need to be cleaned out
			request.addCookie('ctxid', '', path='/')

		# Location
		except responsecodes.HTTPResponseCode, e:
			response = e.code
			if e.msg:
				data = self.router['TemplateRender'](t='/errors/resp', title=e.title or e.__class__.__name__, err=e, reverseinfo=request.uri)
				data = unicode(data).encode('utf-8')
			headers.update(e.headers)

		except IOError, e:
			pass

		# General error
		except BaseException, e:
			data = self.render_error(request.uri, e)

		headers['X-Error'] = e

		[request.setHeader(key, headers[key]) for key in headers]
		request.setResponseCode(response)

		request.write(data)
		request.finish()

		#self._logline(request, size=len(data))



	def _logline(self, request, size=-1, t=None, ctxid=None):
		logdata = dict(
			host = request.getClientIP(),
			ctxid = request.getCookie("ctxid") or '-',
			rtime = time.strftime(emen2.util.loganalyzer.CTIME),
			request = '"%s %s"' % (request.method, request.uri),
			response_code = request.code,
			size = size,
			cputime = 0,
			resource = self.resourcename
		)
		if t:
			logdata['cputime'] = (time.time()-t)*1000

		g.log.msg('WEB', emen2.util.loganalyzer.AccessLogLine(**logdata))



	def cookie_expire_time(self):
		return time.strftime("%a, %d-%b-%Y %H:%M:%S PST", time.localtime(time.time()+604800))



	def render_error(self, location, e):
		args = {'location': location, 'errmsg': str(e)}
		p = emen2.web.views.error.Error()
		p.error(**args)
		data = unicode(p.get_data()).encode("utf-8")
		return data


	def render_security_error(self, location, e):
		args = {'location': location, 'errmsg': str(e)}
		p = emen2.web.views.auth.Auth()
		p.login(**args)
		data = unicode(p.get_data()).encode("utf-8")
		return data














__version__ = "$Revision: 1.33 $".split(":")[1][:-1].strip()
